from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session

from .models.base import Base


def create_sqlalchemy_session(file_path: str) -> Session:
    engine = create_engine(f"sqlite:///{file_path}")
    Base.metadata.create_all(bind=engine)
    return sessionmaker(bind=engine)()



from sqlalchemy.types import String, Integer
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base


class StopTime(Base):
    trip_id: Mapped[str] = mapped_column(String, index=True)
    stop_id: Mapped[str] = mapped_column(String, index=True)
    arrival_time: Mapped[int] = mapped_column(Integer)
    departure_time: Mapped[int] = mapped_column(Integer)
    stop_sequence: Mapped[int] = mapped_column(Integer)
    stop_headsign: Mapped[str] = mapped_column(String)
    pickup_type: Mapped[str] = mapped_column(String)
    drop_off_type: Mapped[str] = mapped_column(String)
    timepoint: Mapped[str] = mapped_column(String)
    checkpoint_id: Mapped[str] = mapped_column(String)
    continuous_pickup: Mapped[str] = mapped_column(String, nullable=True)
    continuous_drop_off: Mapped[str] = mapped_column(String, nullable=True)



from enum import Enum
from typing import Literal

from sqlalchemy.types import String, Integer
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base
from ..utils.enum import gtfs_enum_type


class TransferType(Enum):
    RECOMMENDED = "0"
    TIMED = "1"
    REQUIRES_MINIMUM_TIME = "2"
    NOT_POSSIBLE = "3"
    IN_SEAT = "4"
    MUST_REBOARD = "5"


TTranferType = Literal[
    TransferType.RECOMMENDED,
    TransferType.TIMED,
    TransferType.REQUIRES_MINIMUM_TIME,
    TransferType.NOT_POSSIBLE,
    TransferType.IN_SEAT,
    TransferType.MUST_REBOARD,
]


class WheelchairAccessibility(Enum):
    NO_INFORMATION = "0"
    ACCESSIBLE = "1"
    NOT_ACCESSIBLE = "2"


TWheelchairAccessibility = Literal[
    WheelchairAccessibility.NO_INFORMATION,
    WheelchairAccessibility.ACCESSIBLE,
    WheelchairAccessibility.NOT_ACCESSIBLE,
]


class Transfer(Base):
    from_stop_id: Mapped[str] = mapped_column(String)
    to_stop_id: Mapped[str] = mapped_column(String)
    transfer_type: Mapped[TTranferType] = mapped_column(gtfs_enum_type(TransferType))
    min_transfer_time: Mapped[int] = mapped_column(Integer, nullable=True)
    min_walk_time: Mapped[int] = mapped_column(Integer, nullable=True)
    min_wheelchair_time: Mapped[int] = mapped_column(Integer, nullable=True)
    suggested_buffer_time: Mapped[int] = mapped_column(Integer, nullable=True)
    wheelchair_transfer: Mapped[TWheelchairAccessibility] = mapped_column(
        gtfs_enum_type(WheelchairAccessibility)
    )
    from_trip_id: Mapped[str] = mapped_column(String, nullable=True)
    to_trip_id: Mapped[str] = mapped_column(String, nullable=True)



from enum import Enum
from typing import Literal

from sqlalchemy.types import String, Float
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base
from .routes import RouteType, TRouteType
from ..utils.enum import gtfs_enum_type


class LocationType(Enum):
    STOP = "0"
    STATION = "1"
    ENTRANCE_EXIT = "2"
    GENERIC_NODE = "3"
    BOARDING_AREA = "4"


TLocationType = Literal[
    LocationType.STOP,
    LocationType.STATION,
    LocationType.ENTRANCE_EXIT,
    LocationType.GENERIC_NODE,
    LocationType.BOARDING_AREA,
]


class WheelchairBoardingType(Enum):
    NO_INFORMATION = "0"
    ACCESSIBLE = "1"
    NOT_ACCESSIBLE = "2"


TWheelchairBoardingType = Literal[
    WheelchairBoardingType.NO_INFORMATION,
    WheelchairBoardingType.ACCESSIBLE,
    WheelchairBoardingType.NOT_ACCESSIBLE,
]


class Stop(Base):
    stop_id: Mapped[str] = mapped_column(String)
    stop_code: Mapped[str] = mapped_column(String)
    stop_name: Mapped[str] = mapped_column(String)
    stop_desc: Mapped[str] = mapped_column(String)
    platform_code: Mapped[str] = mapped_column(String, nullable=True)
    platform_name: Mapped[str] = mapped_column(String, nullable=True)
    stop_lat: Mapped[float] = mapped_column(Float, nullable=True)
    stop_lon: Mapped[float] = mapped_column(Float, nullable=True)
    zone_id: Mapped[str] = mapped_column(String)
    stop_address: Mapped[str] = mapped_column(String, nullable=True)
    stop_url: Mapped[str] = mapped_column(String, nullable=True)
    level_id: Mapped[str] = mapped_column(String, nullable=True)
    location_type: Mapped[TLocationType] = mapped_column(
        gtfs_enum_type(LocationType),
        nullable=True,
    )
    parent_station: Mapped[str] = mapped_column(String, index=True)
    wheelchair_boarding: Mapped[TWheelchairBoardingType] = mapped_column(
        gtfs_enum_type(WheelchairBoardingType),
        nullable=True,
    )
    municipality: Mapped[str] = mapped_column(String, nullable=True)
    on_street: Mapped[str] = mapped_column(String, nullable=True)
    at_street: Mapped[str] = mapped_column(String, nullable=True)
    vehicle_type: Mapped[TRouteType] = mapped_column(
        gtfs_enum_type(RouteType), nullable=True
    )



from csv import DictReader
from datetime import date
from typing import Dict, Union, List

import requests

from .utils.time import date_from_string
from .utils.indexes import index_by
from .feed import GtfsFeed

MBTA_GTFS_ARCHIVE_URL = "https://cdn.mbta.com/archive/archived_feeds.txt"


class MbtaGtfsArchive(object):
    _feeds: Dict[str, GtfsFeed]

    def __init__(
        self,
        local_archive_path: str,
        s3_bucket=None,
        archive_url=MBTA_GTFS_ARCHIVE_URL,
    ):
        self.local_archive_path = local_archive_path
        self._s3_bucket = s3_bucket
        self.archive_url = archive_url
        self._load_feeds()

    @property
    def s3_bucket(self):
        if self._s3_bucket is None:
            raise RuntimeError("No S3 bucket configured for archive")
        return self._s3_bucket

    def _load_feeds(self):
        req = requests.get(self.archive_url)
        lines = req.text.splitlines()
        reader = DictReader(lines, delimiter=",")
        feeds = []
        for entry in reader:
            start_date_string = entry["feed_start_date"]
            start_date = date_from_string(start_date_string)
            end_date = date_from_string(entry["feed_end_date"])
            version = entry["feed_version"]
            url = entry["archive_url"]
            gtfs_feed = GtfsFeed(
                archive=self,
                key=start_date_string,
                start_date=start_date,
                end_date=end_date,
                version=version,
                url=url,
            )
            feeds.append(gtfs_feed)
        self._feeds = index_by(feeds, lambda f: f.key)

    def get_feed_by_key(self, key: str):
        return self._feeds.get(key)

    def get_feeds_for_dates(
        self,
        start_date: Union[None, date] = None,
        end_date: Union[None, date] = None,
    ) -> List[GtfsFeed]:
        matching_feeds = []
        for feed in self._feeds.values():
            if feed.matches_date_range(start_date=start_date, end_date=end_date):
                matching_feeds.append(feed)
        return list(sorted(matching_feeds, key=lambda feed: feed.start_date))

    def get_all_feeds(self):
        return self.get_feeds_for_dates()

    def get_feed_for_date(self, target_date: date):
        for feed in self._feeds.values():
            if target_date >= feed.start_date and target_date <= feed.end_date:
                return feed
        return None

    def get_latest_feed(self):
        return self.get_feeds_for_dates()[-1]



from .base import *
from .calendar import *
from .calendar_attributes import *
from .calendar_dates import *
from .feed_info import *
from .lines import *
from .route_patterns import *
from .routes import *
from .shapes import *
from .stop_times import *
from .stops import *
from .transfers import *
from .trips import *



import enum

import sqlalchemy.types


def gtfs_enum_type(vanilla_enum: enum.Enum):
    return sqlalchemy.types.Enum(
        vanilla_enum,
        values_callable=lambda x: [str(member.value) for member in vanilla_enum],
    )



from typing import Callable
from types import GeneratorType
from functools import wraps


def listify(func):
    @wraps(func)
    def new_func(*args, **kwargs):
        r = func(*args, **kwargs)
        if isinstance(r, GeneratorType):
            return list(r)
        else:
            return r

    return new_func



from sqlalchemy.orm import Session
from typing import Dict, Any, Callable, List, Type, Iterable

from .build import GtfsFeedDownloadResult
from .reader import GtfsReader
from .utils.time import date_from_string, seconds_from_string
from .utils.decorators import listify
from .utils.indexes import bucket_by
from .models.base import Base
from .models.calendar_attributes import CalendarAttribute
from .models.calendar_dates import CalendarServiceException
from .models.calendar import CalendarService
from .models.feed_info import FeedInfo
from .models.lines import Line
from .models.route_patterns import RoutePattern
from .models.routes import Route
from .models.shapes import ShapePoint
from .models.stops import Stop
from .models.stop_times import StopTime
from .models.transfers import Transfer
from .models.trips import Trip

RowTransform = Callable[[str], any]
RowTransforms = Dict[str, RowTransform]


def nullable(transform: RowTransform) -> RowTransform:
    def inner_transform(val: str):
        if not val:
            return None
        return transform(val)

    return inner_transform


def transform_row_dict(
    row_dict: Dict[str, str],
    transforms: RowTransforms,
) -> Dict[str, Any]:
    return {
        key: (transforms[key](value) if transforms.get(key) else value)
        for key, value in row_dict.items()
    }


@listify
def get_trip_rows_with_extra_time_fields(
    trip_rows: List[Dict[str, str]],
    stop_time_rows: List[Dict[str, str]],
) -> List[Dict[str, str]]:
    stop_times_by_trip_id = bucket_by(stop_time_rows, "trip_id")
    for trip_row in trip_rows:
        stop_times_for_trip = sorted(
            stop_times_by_trip_id[trip_row["trip_id"]],
            key=lambda stop_time: int(stop_time["stop_sequence"]),
        )
        yield {
            **trip_row,
            "start_time": stop_times_for_trip[0]["arrival_time"],
            "end_time": stop_times_for_trip[-1]["arrival_time"],
        }


def ingest_feed_info(
    session: Session,
    download: GtfsFeedDownloadResult,
    reader: GtfsReader,
) -> FeedInfo:
    num_feed_infos = session.query(FeedInfo).count()
    next_id = 1 + num_feed_infos
    feed_info_dict_raw = next(reader.read_feed_info())
    feed_info_dict = transform_row_dict(
        {
            **feed_info_dict_raw,
            "id": next_id,
            "feed_info_id": next_id,
            "retrieved_from_url": download.url,
            "zip_md5_checksum": download.zip_md5_checksum,
        },
        {
            "feed_start_date": date_from_string,
            "feed_end_date": date_from_string,
        },
    )
    feed_info = FeedInfo(**feed_info_dict)
    session.add(feed_info)
    session.commit()
    return feed_info


def ingest_rows(
    session: Session,
    model: Type[Base],
    feed_info: FeedInfo,
    rows: Iterable[Dict[str, str]],
    transforms: RowTransforms = {},
    individually: bool = False,
):
    mappings = [
        {
            **transform_row_dict(row, transforms),
            "feed_info_id": feed_info.id,
        }
        for row in rows
    ]
    if individually:
        for mapping in mappings:
            session.add(model(**mapping))
    else:
        session.bulk_insert_mappings(model, mappings)


def ingest_gtfs_csv_into_db(
    session: Session,
    download: GtfsFeedDownloadResult,
    reader: GtfsReader,
):
    feed_info = ingest_feed_info(session, download, reader)
    stop_times = list(reader.read_stop_times())
    trips = list(reader.read_trips())
    trip_rows = get_trip_rows_with_extra_time_fields(trips, stop_times)
    ingest_rows(
        session=session,
        model=CalendarService,
        feed_info=feed_info,
        rows=reader.read_calendar(),
        transforms={
            "start_date": date_from_string,
            "end_date": date_from_string,
        },
    )
    ingest_rows(
        session=session,
        model=CalendarServiceException,
        feed_info=feed_info,
        rows=reader.read_calendar_dates(),
        transforms={
            "date": date_from_string,
        },
    )
    ingest_rows(
        session=session,
        model=CalendarAttribute,
        feed_info=feed_info,
        rows=reader.read_calendar_attributes(),
        transforms={
            "rating_start_date": nullable(date_from_string),
            "rating_end_date": nullable(date_from_string),
        },
    )
    ingest_rows(
        session=session,
        model=Line,
        feed_info=feed_info,
        rows=reader.read_lines(),
    )
    ingest_rows(
        session=session,
        model=Route,
        feed_info=feed_info,
        rows=reader.read_routes(),
    )
    ingest_rows(
        session=session,
        model=RoutePattern,
        feed_info=feed_info,
        rows=reader.read_route_patterns(),
    )
    ingest_rows(
        session=session,
        model=ShapePoint,
        feed_info=feed_info,
        rows=reader.read_shapes(),
        transforms={
            "shape_pt_lat": float,
            "shape_pt_lon": float,
            "shape_pt_sequence": int,
            "shape_dist_traveled": nullable(float),
        },
    )
    ingest_rows(
        session=session,
        model=Stop,
        feed_info=feed_info,
        rows=reader.read_stops(),
        transforms={
            "stop_lat": nullable(float),
            "stop_lon": nullable(float),
        },
    )
    ingest_rows(
        session=session,
        model=Transfer,
        feed_info=feed_info,
        rows=reader.read_transfers(),
        transforms={
            "min_transfer_time": nullable(int),
            "min_walk_time": nullable(int),
            "min_wheelchair_time": nullable(int),
            "suggested_buffer_time": nullable(int),
        },
    )
    ingest_rows(
        session=session,
        model=Trip,
        feed_info=feed_info,
        rows=trip_rows,
        transforms={
            "start_time": seconds_from_string,
            "end_time": seconds_from_string,
        },
    )
    ingest_rows(
        session=session,
        model=StopTime,
        feed_info=feed_info,
        rows=stop_times,
        transforms={
            "arrival_time": seconds_from_string,
            "departure_time": seconds_from_string,
            "stop_sequence": int,
        },
    )
    session.commit()



from datetime import date

from sqlalchemy.types import Date, String
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base


class FeedInfo(Base):
    feed_publisher_name: Mapped[str] = mapped_column(String)
    feed_publisher_url: Mapped[str] = mapped_column(String)
    feed_lang: Mapped[str] = mapped_column(String)
    feed_start_date: Mapped[date] = mapped_column(Date)
    feed_end_date: Mapped[date] = mapped_column(Date)
    feed_version: Mapped[str] = mapped_column(String)
    feed_contact_email: Mapped[str] = mapped_column(String, nullable=True)
    feed_contact_url: Mapped[str] = mapped_column(String, nullable=True)
    # These are not part of GTFS
    retrieved_from_url: Mapped[str] = mapped_column(String)
    zip_md5_checksum: Mapped[str] = mapped_column(String)



from typing import Dict, Any, Callable, List, Union


def bucket_by(
    items: List[any],
    key_getter: Union[str, Callable[[Any], str]],
) -> Dict[str, List[any]]:
    res = {}
    if isinstance(key_getter, str):
        key_getter_as_str = key_getter
        key_getter = lambda dict: dict[key_getter_as_str]
    for item in items:
        key = key_getter(item)
        res.setdefault(key, [])
        res[key].append(item)
    return res


def index_by(items: List[any], key_getter: Union[str, Callable[[Any], str]]):
    res = {}
    if isinstance(key_getter, str):
        key_getter_as_str = key_getter
        key_getter = lambda dict: dict[key_getter_as_str]
    for item in items:
        key = key_getter(item)
        res[key] = item
    return res



from enum import Enum
from typing import Literal
import datetime

from sqlalchemy.types import Date, String
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base
from ..utils.enum import gtfs_enum_type


class CalendarServiceExceptionType(Enum):
    ADDED = "1"
    REMOVED = "2"


TCalendarServiceExceptionType = Literal[
    CalendarServiceExceptionType.ADDED,
    CalendarServiceExceptionType.REMOVED,
]


class CalendarServiceException(Base):
    service_id: Mapped[str] = mapped_column(String, index=True)
    date: Mapped[datetime.date] = mapped_column(Date)
    exception_type: Mapped[TCalendarServiceExceptionType] = mapped_column(
        gtfs_enum_type(CalendarServiceExceptionType)
    )
    holiday_name: Mapped[str] = mapped_column(String, nullable=True)



from enum import Enum
from typing import Literal

from sqlalchemy.types import String, Integer
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base
from ..utils.enum import gtfs_enum_type


class RouteType(Enum):
    NOT_SPECIFIED = ""  # Not in GTFS but seen in a lot of Trips
    TRAM = "0"
    METRO = "1"
    RAIL = "2"
    BUS = "3"
    FERRY = "4"
    CABLE_TRAM = "5"
    AERIAL_LIFT = "6"
    FUNICULAR = "7"
    TROLLEYBUS = "11"
    MONORAIL = "12"


TRouteType = Literal[
    RouteType.TRAM,
    RouteType.METRO,
    RouteType.RAIL,
    RouteType.BUS,
    RouteType.FERRY,
    RouteType.CABLE_TRAM,
    RouteType.AERIAL_LIFT,
    RouteType.FUNICULAR,
    RouteType.TROLLEYBUS,
    RouteType.MONORAIL,
]


class Route(Base):
    route_id: Mapped[str] = mapped_column(String)
    agency_id: Mapped[str] = mapped_column(String)
    route_short_name: Mapped[str] = mapped_column(String)
    route_long_name: Mapped[str] = mapped_column(String)
    route_desc: Mapped[str] = mapped_column(String)
    route_type: Mapped[TRouteType] = mapped_column(gtfs_enum_type(RouteType))
    route_url: Mapped[str] = mapped_column(String)
    route_color: Mapped[str] = mapped_column(String)
    route_text_color: Mapped[str] = mapped_column(String)
    route_sort_order: Mapped[int] = mapped_column(Integer)
    route_fare_class: Mapped[str] = mapped_column(String, nullable=True)
    line_id: Mapped[str] = mapped_column(String, index=True, nullable=True)
    listed_route: Mapped[str] = mapped_column(String, nullable=True)



from sqlalchemy.types import String
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base


class Line(Base):
    line_id: Mapped[str] = mapped_column(String)
    line_short_name: Mapped[str] = mapped_column(String)
    line_long_name: Mapped[str] = mapped_column(String)
    line_desc: Mapped[str] = mapped_column(String)
    line_url: Mapped[str] = mapped_column(String)
    line_color: Mapped[str] = mapped_column(String)
    line_text_color: Mapped[str] = mapped_column(String)
    line_sort_order: Mapped[str] = mapped_column(String)



from enum import Enum
from typing import Literal

from sqlalchemy.types import String, Integer
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base
from ..utils.enum import gtfs_enum_type


class RoutePatternTypicality(Enum):
    NOT_DEFINED = "0"
    TYPICAL = "1"
    DEVIATION = "2"
    HIGHLY_ATYPICAL = "3"
    DIVERSION = "4"


TRoutePatternTypicality = Literal[
    RoutePatternTypicality.NOT_DEFINED,
    RoutePatternTypicality.TYPICAL,
    RoutePatternTypicality.DEVIATION,
    RoutePatternTypicality.HIGHLY_ATYPICAL,
    RoutePatternTypicality.DIVERSION,
]


class RoutePattern(Base):
    route_pattern_id: Mapped[str] = mapped_column(String)
    route_id: Mapped[str] = mapped_column(String, index=True)
    direction_id: Mapped[str] = mapped_column(String)
    route_pattern_name: Mapped[str] = mapped_column(String)
    route_pattern_time_desc: Mapped[str] = mapped_column(String)
    route_pattern_typicality: Mapped[TRoutePatternTypicality] = mapped_column(
        gtfs_enum_type(RoutePatternTypicality)
    )
    route_pattern_sort_order: Mapped[int] = mapped_column(Integer)
    representative_trip_id: Mapped[str] = mapped_column(String)



from enum import Enum
from typing import Literal

from sqlalchemy.types import String, Integer
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base
from .routes import RouteType, TRouteType
from ..utils.enum import gtfs_enum_type


class WheelchairAccessibility(Enum):
    NO_INFORMATION = "0"
    ACCESSIBLE = "1"
    NOT_ACCESSIBLE = "2"


TWheelchairAccessibility = Literal[
    WheelchairAccessibility.NO_INFORMATION,
    WheelchairAccessibility.ACCESSIBLE,
    WheelchairAccessibility.NOT_ACCESSIBLE,
]


class BikesAllowed(Enum):
    NO_INFORMATION = "0"
    ALLOWED = "1"
    NOT_ALLOWED = "2"


TBikesAllowed = Literal[
    BikesAllowed.NO_INFORMATION,
    BikesAllowed.ALLOWED,
    BikesAllowed.NOT_ALLOWED,
]


class Trip(Base):
    route_id: Mapped[str] = mapped_column(String, index=True)
    service_id: Mapped[str] = mapped_column(String, index=True)
    trip_id: Mapped[str] = mapped_column(String)
    trip_headsign: Mapped[str] = mapped_column(String)
    trip_short_name: Mapped[str] = mapped_column(String)
    direction_id: Mapped[str] = mapped_column(String)
    block_id: Mapped[str] = mapped_column(String)
    shape_id: Mapped[str] = mapped_column(String)
    wheelchair_accessible: Mapped[TWheelchairAccessibility] = mapped_column(
        gtfs_enum_type(WheelchairAccessibility)
    )
    trip_route_type: Mapped[TRouteType] = mapped_column(gtfs_enum_type(RouteType))
    route_pattern_id: Mapped[str] = mapped_column(String, nullable=True)
    bikes_allowed: Mapped[TBikesAllowed] = mapped_column(
        gtfs_enum_type(BikesAllowed),
        nullable=True,
    )
    # start_time and end_time are not part of GTFS but they're useful information to have
    # in smaller versions of the database without StopTimes
    start_time: Mapped[int] = mapped_column(Integer)
    end_time: Mapped[int] = mapped_column(Integer)



from dataclasses import dataclass
from os import path, remove
from shutil import copy
from zipfile import ZipFile
from hashlib import md5


import requests

from .reader import GtfsReader
from .feed import GtfsFeed


@dataclass
class GtfsFeedDownloadResult(object):
    url: str
    zip_md5_checksum: str


def download_feed_zip(feed_url: str, target_path: str):
    response = requests.get(feed_url, stream=True)
    block_size = 1024
    with open(target_path, "wb") as file:
        for data in response.iter_content(block_size):
            file.write(data)


def unzip_feed(zip_path: str, target_path: str):
    zf = ZipFile(zip_path)
    zf.extractall(target_path)


def get_zip_checksum(zip_path: str) -> str:
    with open(zip_path, "rb") as file:
        chunk_size = 4096
        hasher = md5()
        while chunk := file.read(chunk_size):
            hasher.update(chunk)
    checksum = hasher.hexdigest()
    return checksum


def ingest_feed_to_sqlite(
    feed_path: str,
    db_path: str,
    compact_db_path: str,
    result: GtfsFeedDownloadResult,
):
    from .ingest import ingest_gtfs_csv_into_db
    from .compact import make_compact_db
    from .session import create_sqlalchemy_session

    try:
        reader = GtfsReader(feed_path)
        session = create_sqlalchemy_session(db_path)
        ingest_gtfs_csv_into_db(session, result, reader)
        copy(db_path, compact_db_path)
        compact_session = create_sqlalchemy_session(compact_db_path)
        make_compact_db(compact_session)
    except Exception as ex:
        try:
            remove(db_path)
            remove(compact_db_path)
        except FileNotFoundError:
            pass
        raise ex


def build_local_feed_entry(feed: GtfsFeed, compact_only=False):
    (zip_path, feed_path, db_path, compact_db_path) = (
        path.join(feed.local_subdirectory, entity)
        for entity in ("data.zip", "feed", "gtfs.sqlite3", "gtfs_compact.sqlite3")
    )
    download_feed_zip(feed.url, zip_path)
    unzip_feed(zip_path, feed_path)
    result = GtfsFeedDownloadResult(
        url=feed.url,
        zip_md5_checksum=get_zip_checksum(zip_path),
    )
    ingest_feed_to_sqlite(feed_path, db_path, compact_db_path, result)
    if compact_only:
        remove(db_path)



from enum import Enum
from datetime import date
from typing import Literal

from sqlalchemy.types import Date, String
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base
from ..utils.enum import gtfs_enum_type


class ServiceDayAvailability(Enum):
    NOT_AVAILABLE = "0"
    AVAILABLE = "1"


TServiceDayAvailability = Literal[
    ServiceDayAvailability.AVAILABLE,
    ServiceDayAvailability.NOT_AVAILABLE,
]


class CalendarService(Base):
    service_id: Mapped[str] = mapped_column(String)
    monday: Mapped[TServiceDayAvailability] = mapped_column(
        gtfs_enum_type(ServiceDayAvailability)
    )
    tuesday: Mapped[TServiceDayAvailability] = mapped_column(
        gtfs_enum_type(ServiceDayAvailability)
    )
    wednesday: Mapped[TServiceDayAvailability] = mapped_column(
        gtfs_enum_type(ServiceDayAvailability)
    )
    thursday: Mapped[TServiceDayAvailability] = mapped_column(
        gtfs_enum_type(ServiceDayAvailability)
    )
    friday: Mapped[TServiceDayAvailability] = mapped_column(
        gtfs_enum_type(ServiceDayAvailability)
    )
    saturday: Mapped[TServiceDayAvailability] = mapped_column(
        gtfs_enum_type(ServiceDayAvailability)
    )
    sunday: Mapped[TServiceDayAvailability] = mapped_column(
        gtfs_enum_type(ServiceDayAvailability)
    )
    start_date: Mapped[date] = mapped_column(Date)
    end_date: Mapped[date] = mapped_column(Date)



import datetime


def date_from_string(date_str: str):
    return datetime.datetime.strptime(date_str, "%Y%m%d").date()


def date_to_string(date: datetime.date):
    return date.strftime("%Y%m%d")


def timedelta_from_string(time_str: str):
    pieces = [int(x) for x in time_str.split(":")]
    if len(pieces) == 3:
        hours, minutes, seconds = pieces
        return datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
    hours, minutes = pieces
    return datetime.timedelta(hours=hours, minutes=minutes)


def seconds_from_string(time_str: str):
    td = timedelta_from_string(time_str)
    return int(td.total_seconds())



from sqlalchemy import text
from sqlalchemy.orm import Session

from .models.stop_times import StopTime


def vacuum(session: Session):
    session.commit()
    session.execute(text("vacuum"))
    session.commit()


def make_compact_db(session: Session):
    session.query(StopTime).delete()
    vacuum(session)



from typing import TYPE_CHECKING, Union
from os import path, mkdir, remove, rmdir
from datetime import date
from dataclasses import dataclass
from functools import cached_property


if TYPE_CHECKING:
    from .archive import MbtaGtfsArchive

DB_FILE = "gtfs.sqlite3"
DB_COMPACT_FILE = "gtfs_compact.sqlite3"
ALL_DB_FILES = [DB_FILE, DB_COMPACT_FILE]


@dataclass
class GtfsFeed(object):
    archive: "MbtaGtfsArchive"
    key: str
    url: str
    version: str
    start_date: date
    end_date: date

    def __post_init__(self):
        self.compact_only = False

    def __repr__(self):
        return f"GtfsFeed({self.key})"

    @cached_property
    def local_subdirectory(self):
        return path.join(self.archive.local_archive_path, self.key)

    def required_feed_files(self):
        return ALL_DB_FILES if not self.compact_only else [DB_COMPACT_FILE]

    def use_compact_only(self, enabled: bool = True):
        self.compact_only = enabled

    def ensure_subdirectory(self):
        if not path.exists(self.archive.local_archive_path):
            mkdir(self.archive.local_archive_path)
        if not path.exists(self.local_subdirectory):
            mkdir(self.local_subdirectory)

    def exists_locally(self):
        for file in self.required_feed_files():
            if not path.exists(path.join(self.local_subdirectory, file)):
                return False
        return True

    def exists_remotely(self):
        try:
            remote_objects = self.archive.s3_bucket.objects.filter(Prefix=self.key)
        except RuntimeError:
            return False
        for file in self.required_feed_files():
            if not any(obj.key.endswith(file) for obj in remote_objects):
                return False
        return True

    def build_locally(self):
        from .build import build_local_feed_entry

        self.ensure_subdirectory()
        build_local_feed_entry(self, self.compact_only)

    def download_from_s3(self):
        if not self.exists_remotely():
            raise RuntimeError("Feed does not exist remotely")
        self.ensure_subdirectory()
        for file in self.required_feed_files():
            self.archive.s3_bucket.download_file(
                f"{self.key}/{file}", path.join(self.local_subdirectory, file)
            )

    def upload_to_s3(self):
        if not self.exists_locally():
            raise RuntimeError("Feed does not exist locally")
        for file in self.required_feed_files():
            self.archive.s3_bucket.upload_file(
                path.join(self.local_subdirectory, file), f"{self.key}/{file}"
            )

    def download_or_build(self):
        if self.exists_locally():
            return
        if self.exists_remotely():
            self.download_from_s3()
            return
        self.build_locally()

    def delete_locally(self):
        if not self.exists_locally():
            raise RuntimeError("Feed does not exist locally")
        for file in self.required_feed_files():
            file_path = path.join(self.local_subdirectory, file)
            if path.exists(file_path):
                remove(file_path)
        # Delete self.local_subdirectory if empty
        try:
            rmdir(self.local_subdirectory)
        except OSError:
            pass

    def create_sqlite_session(self, compact=False):
        if not self.exists_locally():
            raise RuntimeError("Feed does not exist locally")
        from .session import create_sqlalchemy_session

        db_path = path.join(
            self.local_subdirectory,
            "gtfs_compact.sqlite3" if compact else "gtfs.sqlite3",
        )
        return create_sqlalchemy_session(db_path)

    def matches_date_range(
        self,
        start_date: Union[None, date],
        end_date: Union[None, date],
    ):
        if start_date is None and end_date is None:
            return True
        if start_date is None:
            return end_date >= self.start_date
        if end_date is None:
            return self.end_date >= start_date
        return end_date >= self.start_date and self.end_date >= start_date



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mbta_gtfs_sqlite', 'mbta_gtfs_sqlite.models', 'mbta_gtfs_sqlite.utils']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=2.0.7,<3.0.0', 'requests>=2.28.2,<3.0.0']

setup_kwargs = {
    'name': 'mbta-gtfs-sqlite',
    'version': '0.9.4',
    'description': "Query the MBTA's static GTFS feeds using sqlite",
    'long_description': None,
    'author': 'Ian Reynolds',
    'author_email': 'ireynolds@transitmatters.info',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from sqlalchemy.orm import DeclarativeBase, mapped_column
from sqlalchemy import Integer, ForeignKey

from sqlalchemy.ext.declarative import declared_attr


class Base(DeclarativeBase):
    # Automatically infer each table name from the class name
    @declared_attr
    def __tablename__(cls):
        return cls.__name__

    # Provide a unique ID
    id = mapped_column(Integer, autoincrement=True, primary_key=True)

    # Basically every model is going to belong to a specific GtfsFeedInstance
    feed_info_id = mapped_column(ForeignKey("FeedInfo.id"), nullable=False, index=True)



from enum import Enum
from typing import Literal
from datetime import date

from sqlalchemy.types import String, Date
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base
from ..utils.enum import gtfs_enum_type


class ServiceScheduleTypicality(Enum):
    NOT_DEFINED = "0"
    MINOR_MODIFICATIONS = "1"
    EXTRA_SERVICE = "2"
    REDUCED_TO_WEEKEND_SERVICE = "3"
    MAJOR_PLANNED_DISRUPTION = "4"
    MAJOR_ATYPICAL_REDUCTIONS = "5"


TServiceScheduleTypicality = Literal[
    ServiceScheduleTypicality.NOT_DEFINED,
    ServiceScheduleTypicality.MINOR_MODIFICATIONS,
    ServiceScheduleTypicality.EXTRA_SERVICE,
    ServiceScheduleTypicality.REDUCED_TO_WEEKEND_SERVICE,
    ServiceScheduleTypicality.MAJOR_PLANNED_DISRUPTION,
    ServiceScheduleTypicality.MAJOR_ATYPICAL_REDUCTIONS,
]


class CalendarAttribute(Base):
    service_id: Mapped[str] = mapped_column(String, index=True)
    service_description: Mapped[str] = mapped_column(String)
    service_schedule_name: Mapped[str] = mapped_column(String)
    service_schedule_type: Mapped[str] = mapped_column(String)
    service_schedule_typicality: Mapped[TServiceScheduleTypicality] = mapped_column(
        gtfs_enum_type(ServiceScheduleTypicality)
    )
    rating_start_date: Mapped[date] = mapped_column(Date, nullable=True)
    rating_end_date: Mapped[date] = mapped_column(Date, nullable=True)
    rating_description: Mapped[str] = mapped_column(String, nullable=True)



from csv import DictReader
from os import path


class GtfsReader:
    def _reader_by_file_name(self, name: str):
        file_path = path.join(self.root, name + ".txt")

        def load():
            try:
                with open(file_path, "r") as file:
                    dict_reader = DictReader(file)
                    for row in dict_reader:
                        yield row
            except FileNotFoundError:
                yield from []

        return load

    def __init__(self, root: str):
        self.root = root
        self.read_calendar = self._reader_by_file_name("calendar")
        self.read_calendar_attributes = self._reader_by_file_name("calendar_attributes")
        self.read_calendar_dates = self._reader_by_file_name("calendar_dates")
        self.read_feed_info = self._reader_by_file_name("feed_info")
        self.read_lines = self._reader_by_file_name("lines")
        self.read_route_patterns = self._reader_by_file_name("route_patterns")
        self.read_routes = self._reader_by_file_name("routes")
        self.read_shapes = self._reader_by_file_name("shapes")
        self.read_stop_times = self._reader_by_file_name("stop_times")
        self.read_stops = self._reader_by_file_name("stops")
        self.read_transfers = self._reader_by_file_name("transfers")
        self.read_trips = self._reader_by_file_name("trips")



from sqlalchemy.types import String, Integer, Float
from sqlalchemy.orm import mapped_column, Mapped

from .base import Base


class ShapePoint(Base):
    shape_id: Mapped[str] = mapped_column(String, index=True)
    shape_pt_lat: Mapped[float] = mapped_column(Float)
    shape_pt_lon: Mapped[float] = mapped_column(Float)
    shape_pt_sequence: Mapped[int] = mapped_column(Integer)
    shape_dist_traveled: Mapped[float] = mapped_column(Float, nullable=True)
